package ucl.cities.context;

import gov.nasa.worldwind.render.Polyline;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.measure.unit.*;
import javax.vecmath.Point3d;
import javax.measure.quantity.Length;

import org.jscience.geography.coordinates.LatLong;
import org.jscience.geography.coordinates.UTM;
import org.jscience.geography.coordinates.crs.ReferenceEllipsoid;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.Point;

import repast.simphony.context.Context;
import repast.simphony.context.space.continuous.ContinuousSpaceFactoryFinder;
import repast.simphony.context.space.gis.GeographyFactoryFinder;
import repast.simphony.context.space.graph.NetworkFactoryFinder;
import repast.simphony.context.space.grid.GridFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.continuous.RandomCartesianAdder;
import repast.simphony.space.continuous.StickyBorders;
import repast.simphony.space.gis.Geography;
import repast.simphony.space.gis.GeographyParameters;
import repast.simphony.space.graph.Network;
import repast.simphony.space.grid.GridBuilderParameters;
import repast.simphony.space.grid.RandomGridAdder;
import ucl.cities.model.NewWilsonModel;
import ucl.cities.model.WilsonAgentModel;
import ucl.cities.output.CityOutputter;
import ucl.cities.output.Outputter;
import ucl.cities.output.PajekOutputter;
import ucl.cities.utilities.DistanceMeasure;
import ucl.cities.dataLoad.ShapeFileLoader;
import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.City;
import ucl.cities.entities.Location;

/**
 * Simulation context builder.
 * 
 * @author maltaweel
 *
 * @param <T>
 * @param <Length>
 */
public class CitiesContextBuilder<T, Length> implements ContextBuilder<T>{

	private Network<Location> cNetwork;  //the canal network
//	private Network<T> hydrologyNetwork;
//	private Geography geography;
//	private static boolean started=false;
	List<CanalEdge>edges=new ArrayList<CanalEdge>();
	private Geography geography;
	public Map<String,Object>canalPtsData = new HashMap<String,Object>();
	public Map<String,Object>edgeData = new HashMap<String,Object>();
	private Boolean donkeys;
	public static List<City>cities=new ArrayList<City>();
	Map<City,Map> cityMap = new HashMap<City,Map>();
	
	@Override
	public Context build(Context context) {
		Outputter.runNumber=RunState.getInstance().getRunInfo().getRunNumber();
		Outputter.started=false;
		Parameters p = RunEnvironment.getInstance().getParameters();
		GeographyParameters geoParams = new GeographyParameters();
		  geography = GeographyFactoryFinder.createGeographyFactory(null)
     	   .createGeography("Geography", context, geoParams);
		  
		double endTick=((Double)p.getValue("endTick")).intValue();
		double e1=((Double)p.getValue("epsilon1")).doubleValue();
		double c=((Double)p.getValue("constant")).intValue();
//		double e2=((Double)p.getValue("epsilon2")).doubleValue();
//		double e3=((Double)p.getValue("epsilon3")).doubleValue();
		
		NewWilsonModel.Epsilon1=e1;
		NewWilsonModel.Constant=c;
//		WilsonAgentModel.Epsilon2=e2;
//		WilsonAgentModel.Epsilon3=e3;
		
		Network<Location> canalNetwork = 
			(Network<Location>) NetworkFactoryFinder.createNetworkFactory(null).
			createNetwork("Canals",context,false);
		
		Network<Location> flowNetwork = 
			(Network<Location>) NetworkFactoryFinder.createNetworkFactory(null).
			createNetwork("Flow",context,true);
		
		Network<Location> ndNetwork = 
			(Network<Location>) NetworkFactoryFinder.createNetworkFactory(null).
			createNetwork("nd",context,true);
		
		GridFactoryFinder.createGridFactory(null).createGrid("Simple Grid",//
				context, GridBuilderParameters.singleOccupancy2DTorus(new RandomGridAdder<Location>(),
								 1000000, 3000000));
				double[] size = { 1000000, 1000000, 1000000 };
		ContinuousSpaceFactoryFinder.createContinuousSpaceFactory(null)
				.createContinuousSpace("Simple Space", context, new RandomCartesianAdder<Location>(),
								new repast.simphony.space.continuous.StickyBorders(), size);
		
		this.cNetwork=canalNetwork;
		
		String canalLoc=p.getValueAsString("canalData");
		String canalPts=p.getValueAsString("canalPts");
		String fileLoc=p.getValueAsString("scenarioData");
		this.donkeys=(Boolean)p.getValue("donkeys");
		
	//	loadCanalPts(context,canalPts);
		loadCities(context,fileLoc);
		loadCanals();
		
		RunEnvironment.getInstance().endAt(endTick);
		
//		if (!RunEnvironment.getInstance().isBatch()){
			CityOutputter co = new CityOutputter();
			context.add(co);
//		}
			
		PajekOutputter po = new PajekOutputter(endTick);
		context.add(po);
		
		return context;
	}
	
	/**
	 * Method to load city data and create cities.
	 * @param context the context to add cities to.
	 * @param fileLoc the location of the shapefile for cities.
	 */
	public void loadCities(Context context, String fileLoc){
		
		 cities.clear();
		 
		 List<Map<String,Object>> data=ShapeFileLoader.shpLoader(new File(fileLoc));
		 Parameters par = RunEnvironment.getInstance().getParameters();
		 
		 Double beta=((Float)par.getValue("beta")).doubleValue();
	//	 double alpha = ((Double)par.getValue("alpha")).doubleValue();
		 ReferenceEllipsoid re=ReferenceEllipsoid.WGS84; 
		 for(Map m : data){
			 Point p=(Point) m.get("the_geom");
			 double x=p.getX();
			 double y=p.getY();
			 String name=(String)m.get("TellSite");
			 City c=null;
			 if(name.contains("Bakr 'Awa") ||name.contains("Yasin Tape (SSP-002)")){
				 c = new City();
			 }
			 else{
				 double rr=RandomHelper.nextDouble();
				if(rr>0.46)
					continue;
				c = new City();
			 }
			 double alpha = ((Double)m.get("alpha")).doubleValue();
			 String number=m.get("cat").toString();
			 Integer z=(Integer)m.get("Z");
	//		 Long pop=(Long)m.get("Population");
			 Long pop=(Long)m.get("Population");
			 double size=(Double)m.get("Size");
			 double scalars=(Double)m.get("Size2");
			 double external=(Double)m.get("externalFL");
			 
			 c.setGeo(p);
			 c.setOrigX(x);
			 c.setOrigY(y);
			 UTM utm = UTM.valueOf(38, 'N', x, y,SI.METER);
			 LatLong ll=UTM.utmToLatLong(utm, re);
			 double ccs=ll.latitudeValue(NonSI.DEGREE_ANGLE);
			 double ccs2=ll.longitudeValue(NonSI.DEGREE_ANGLE);
			 c.setX(ccs);
			 c.setY(ccs2);
			 c.setZ(z);
			 c.setPoint(p);
			 c.externalFlowPercentage=external;
			 c.setNumber(Integer.valueOf(number));
			 if(name!=null && !name.equals(" ")){
				c.setName(name);
			 }
			 else
				 c.setName(number);
	//		 determineRoadPoint(c);
			 c.setModel(new NewWilsonModel());
			 c.setBeta((Double)beta);
			 c.setAlpha(alpha);
			 c.setSize(size);
			 c.setPopulation(pop.intValue());
			 c.setScalars(scalars);
			 context.add(c);
			 geography.move(c, p);
			 cityMap.put(c, m);
			 cities.add(c);
		 }
	}
	
	/**
	 * Method to load canal data.
	 * @param <Location>
	 * @param context the context.
	 * @param loc the location of the file.
	 */
/*	public void loadCanals(Context context, String loc){
		List<Map<String,Object>> data=ShapeFileLoader.shpLoader(new File(loc));
		 ReferenceEllipsoid re=ReferenceEllipsoid.WGS84; 
		 for(Map m : data){
			 MultiLineString mls=(MultiLineString) m.get("the_geom");
			 double weight=(Double)m.get("weight"); //for weight (i.e., number of crossings with 1 as default then 1+ for each crossing)
			 double rho=(Double)m.get("rho");  //for width of traversed rivers if needed
			 for(int i=0; i < mls.getNumGeometries(); i++){
				 LineString g=(LineString) mls.getGeometryN(i);
				 for(int t=0; t < g.getNumPoints()-1; t++){
					 Point p1=g.getPointN(t);
					 Point p2=g.getPointN(t+1);
					 double x=Math.round(p1.getX());
					 double y=Math.round(p1.getY());
					 if(p1.getX()==p2.getX() && p2.getX()==p2.getX())
						 continue;
					 double z=0;
					 String xy=Double.toString(Math.rint(x))+","+Double.toString(Math.rint(y));
					 Location l=null;
				 
					 if(canalPtsData.containsKey(xy)){
						 l=(Location) canalPtsData.get(xy);
					 }
					 else{
						 UTM utm = UTM.valueOf(38, 'N', x, y,SI.METER);
						 LatLong ll=UTM.utmToLatLong(utm, re);
						 double ccs=ll.latitudeValue(NonSI.DEGREE_ANGLE);
						 double ccs2=ll.longitudeValue(NonSI.DEGREE_ANGLE);
						 l=new Location(ccs,ccs2,z);
						 context.add(l);
					 	 canalPtsData.put(xy,l);
					 }
					 double x2=p2.getX();
					 double y2=p2.getY();
					 String xy2=Double.toString(Math.rint(x2))+","+Double.toString(Math.rint(y2));
					 Location l2=null;
					 if(canalPtsData.containsKey(xy2)){
						 l2=(Location) canalPtsData.get(xy2);
					 }
					 else{
						 UTM utm = UTM.valueOf(38, 'N', x2, y2,SI.METER);
						 LatLong ll=UTM.utmToLatLong(utm, re);
						 double ccs=ll.latitudeValue(NonSI.DEGREE_ANGLE);
						 double ccs2=ll.longitudeValue(NonSI.DEGREE_ANGLE);
						 l2=new Location(ccs,ccs2,z);
						 canalPtsData.put(xy2,l2);
						 context.add(l2);
					 }
				  
				 CanalEdge<Location> e=null;
				 if(edgeData.containsKey(l.getX()+","+l.getY()+","
						 +l2.getX()+","+l2.getY()) || 
						 edgeData.containsKey(l2.getX()+","+l2.getY()+","
								 +l.getX()+","+l.getY())){
					 e=(CanalEdge<Location>)edgeData.get(l.getX()+","+l.getY()+","
						 +l2.getX()+","+l2.getY());
					 if(e==null){
						 e=(CanalEdge<Location>)edgeData.get(l2.getX()+","+l2.getY()+","
								 +l.getX()+","+l.getY());
					 }
				 }
				 else{
					 e = new CanalEdge(l,l2);
					 
					 if(!donkeys){
						 DistanceMeasure.notDonkeyEnergy(p1,p2, l,l2,e);
					 }
					 else{
						 e.setWeight(rho*weight*DistanceMeasure.calculateDonkeyEnergy(p1, p2,l,l2));
						 e.setrWeight(rho*weight*DistanceMeasure.calculateDonkeyEnergy(p2, p1,l,l2));
					 }
					 edgeData.put(l.getX()+","+l.getY()+","
							 +l2.getX()+","+l2.getY(),e);
				 }
				 if(!cNetwork.containsEdge(e)){					 
					 cNetwork.addEdge(e);
					 edges.add(e);
				 }
//				 Polyline line = new Polyline();
				 geography.move(e,g);
				}
				 
			 }		 
		 }
	} */
	
	/**
	 * Weights between cities (could be roads).
	 */
	public void loadCanals(){
		int s = cities.size();
		 for(int i=0; i < s; i++){
			 City c = cities.get(i);
			 int n=c.getNumber();
			 Map m = cityMap.get(c);
			 Point p=c.getPoint();
			 cities.remove(c);
			 Iterator<City> ic2 = cities.iterator();
			 while(ic2.hasNext()){
				 City c2 = ic2.next();
				 int n2=c2.getNumber();
				 double v=(Double)m.get("point"+n2);
	//			 double ccc=(Double)m.get("crossing"+n2);
				 CanalEdge<Location> e=new CanalEdge(c,c2);
				 if(!donkeys){
					 DistanceMeasure.notDonkeyEnergy(c.point,c2.point, c,c2,e);
				 }
				 else{
					 e.setWeight(v);
				//	 e.setrWeight(rho*weight*DistanceMeasure.calculateDonkeyEnergy(c.point,c2.point, c,c2));
				 }
		//		 e.setCrossingWeight(ccc);
				 if(!cNetwork.containsEdge(e)){					 
					 cNetwork.addEdge(e);
					 edges.add(e);
				 }
			 }
			 cities.add(i,c);
//				 Polyline line = new Polyline();
//				 geography.move(e,g);
		 }
	}
	/**
	 * Method to load canal pts.
	 * @param context the context.
	 * @param loc the location.
	 */
	public void loadCanalPts(Context context, String loc){
		List<Map<String,Object>> data=ShapeFileLoader.shpLoader(new File(loc));
		int i=0;
		 ReferenceEllipsoid re=ReferenceEllipsoid.WGS84; 
		 for(Map m : data){
			 Point p=(Point) m.get("the_geom");
			 double x=p.getX();
			 double y=p.getY();
			 UTM utm = UTM.valueOf(38, 'N', x, y,SI.METER);
			 LatLong ll=UTM.utmToLatLong(utm, re);
			 String name=(String)m.get("NAME");
			 
			 Long z = (Long) m.get("Z2");
			 double ccs=ll.latitudeValue(NonSI.DEGREE_ANGLE);
			 double ccs2=ll.longitudeValue(NonSI.DEGREE_ANGLE);
			 Location l = new Location(ccs,ccs2,z);
			 l.point=p;
			 if(name!=null){
				 l.setName(name);
			 }
			 canalPtsData.put(Double.toString(Math.rint(x))+","
					 +Double.toString(Math.rint(y)), l);
			 context.add(l);
			 geography.move(l, p);
			 i++;
		 }
	}
	
	/**
	 * Find the nearest road point on the canal network.
	 * @param c a city object.
	 */
	public void determineRoadPoint(City c){
		Location l = null;
		Iterator<CanalEdge> ci=edges.iterator();
		
		double distance=0;
		
		while(ci.hasNext()){
			CanalEdge<Location> e = (CanalEdge<Location>)ci.next();
			Location ll=(Location)e.getSource();
			Location ll2=(Location)e.getTarget();
			
			Point3d p3d = new Point3d(c.getOrigX(),c.getOrigY(),c.getZ());
			double d1=p3d.distance(new Point3d(ll.point.getX(),ll.point.getY(),ll.getZ()));
			double d2=p3d.distance(new Point3d(ll2.point.getX(),ll2.point.getY(),ll2.getZ()));
			
			if(distance==0){
				distance=d1;
				l=ll;
			}
			
			if(distance>d1){
				l=ll;
				distance=d1;
			}
			if(distance>d2){
				l=ll2;
				distance=d2;
			}
		}
		c.setRoadPointLocation(l);
	}
}
