package ucl.cities.entities;

import gov.nasa.worldwind.awt.WorldWindowGLCanvas;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.layers.IconLayer;
import gov.nasa.worldwind.layers.RenderableLayer;
import gov.nasa.worldwind.render.Polyline;
import gov.nasa.worldwind.render.Renderable;
import gov.nasa.worldwind.render.UserFacingIcon;

import java.awt.Color;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.gisnew.AbstractSurfaceLayer;
import repast.simphony.visualization.gisnew.NewDisplayGIS;
import ucl.cities.model.Model;
import ucl.cities.model.WilsonAgentModel;
import ucl.cities.output.Outputter;

/**
 * The city entity.
 * 
 * @author maltaweel
 *
 */
public class City extends Location{

	public double size;   //size of the city
	public int number;  //nubmer in the survey
	public double alpha ;  //the city's alpha value
	public double beta;  //the city's beta
	public double lambda;  // the city's lambda
	public double h;   //city's h
	public double s;  //the city's s
	public double delta;  //the delta 
	public boolean interactSize;  //does interact in size
	public double price;  // the price of something for interaction
    public double revenue;  //revenue earned
    public double rent;  // the rent
    public double scalars; //scalar expenditure value
    public double flowValue=0d; //the value of flow in the entropy model
    public double externalFlow=0d;  //external flow into city
    public double externalFlowPercentage;  // percent of flow from external sources
    public Location roadPointLocation;  //nearest road point location
    public int population;  //population of the city
    public double pop;
    public double pop_sd;
    public double flow_sd;
    public Model model;  // the Wilson model
    public Point point;  //point location
	private UserFacingIcon icon;
	private Map<RepastEdge,RenderableLayer> oldRend=new HashMap<RepastEdge,RenderableLayer>();
	private IconLayer iconLayer;
    private static boolean didIt=false;
    Map<String,Double>costs=new HashMap<String,Double>();  //cost of travel between sites
    double volumeFromI;  //volume of flow from I
    
	/**
	 * General constructor
	 */
	public City(){
		
	}

	
	public double getAlpha() {
		return alpha;
	}

	public int getPopulation() {
		return population;
	}

	public void setPopulation(int population) {
		this.population = population;
	}

    
	public Point getPoint() {
		return point;
	}

	
	
	public int getNumber() {
		return number;
	}


	public void setNumber(int number) {
		this.number = number;
	}


	public void setPoint(Point point) {
		this.point = point;
	}


	public void setAlpha(double alpha) {
		this.alpha = alpha;
	}

	public double getSize() {
		return size;
	}

	public void setSize(double size) {
		this.size = size;
	}

	public double getDelta() {
		return delta;
	}

	public void setDelta(double delta) {
		this.delta = delta;
	}
	
	
	public boolean isInteractSize() {
		return interactSize;
	}

	public void setInteractSize(boolean interactSize) {
		this.interactSize = interactSize;
	}

	
	public double getBeta() {
		return beta;
	}

	public void setBeta(double beta) {
		this.beta = beta;
	}

	public double getLambda() {
		return lambda;
	}

	public void setLambda(double lambda) {
		this.lambda = lambda;
	}

	public double getH() {
		return h;
	}

	public void setH(double h) {
		this.h = h;
	}

	public double getS() {
		return s;
	}

	public void setS(double s) {
		this.s = s;
	}
	
	

	public double getPrice() {
		return price;
	}

	public void setPrice(double price) {
		this.price = price;
	}

	
	public double getRevenue() {
		return revenue;
	}

	public void setRevenue(double revenue) {
		this.revenue = revenue;
	}

	public double getRent() {
		return rent;
	}

	public void setRent(double rent) {
		this.rent = rent;
	}
	
	

	public double getScalars() {
		return scalars;
	}

	public void setScalars(double scalars) {
		this.scalars = scalars;
	}

	
	public double getFlowValue() {
		return flowValue;
	}

	public void setFlowValue(double flowValue) {
		this.flowValue = flowValue;
	}

	
	public Model getModel() {
		return model;
	}


	public void setModel(Model model) {
		this.model = model;
	}

	
	public Location getRoadPointLocation() {
		return roadPointLocation;
	}


	public void setRoadPointLocation(Location roadPointLocation) {
		this.roadPointLocation = roadPointLocation;
	}

	@ScheduledMethod(start = 1, interval=1, pick=ScheduledMethod.ALL, priority = ScheduleParameters.FIRST_PRIORITY)
	public void runFirstClear(){
		this.setFlowValue(0);
	}
	
	@ScheduledMethod(start = 1, interval=1, pick=ScheduledMethod.ALL, priority = ScheduleParameters.FIRST_PRIORITY-1)
	public void runModel(){
		model.step(this);
	}
	
/*	@ScheduledMethod(start = 10, interval=20,pick=ScheduledMethod.ALL, priority = ScheduleParameters.FIRST_PRIORITY-1)
	public void evolutionModel(){
		double r=RandomHelper.nextDouble();
		if(this.name.equals("Tell Dlehim")){
			this.externalFlowPercentage=2.0;
			this.alpha=5.0;
		}
		else{
			this.externalFlowPercentage=1.0;
			this.alpha=1.0;
		}
			
			
	}  */
	
	@ScheduledMethod(start = 0, interval=1, pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY+1)
	public void runSecondStep(){
		model.calculateSize(this);
//		List<String>output=new ArrayList<String>();
		Outputter.doOutput(this);
		System.out.println(this.name+" : "+RunEnvironment.getInstance().getCurrentSchedule().getTickCount()+ " : "+flowValue);
	}
	
//	@ScheduledMethod(start = 1, interval=1,pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
	public void initV(){
	 List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
	 Iterator<IDisplay> ic = displays.iterator();
	 while(ic.hasNext()){
		 IDisplay display = ic.next();
		 if(display instanceof NewDisplayGIS){
			 if(gis==null)
				 gis = (NewDisplayGIS)display;
			 WorldWindowGLCanvas canvas=gis.getWwglCanvas();
			 
				Map<Class, AbstractSurfaceLayer> objs=gis.getClassStyleMap();
				AbstractSurfaceLayer layer = objs.get(City.class);
				Collection<Renderable> renderables=(Collection<Renderable>)layer.getRenderables();
				Iterator<Renderable> ico=layer.getRenderables().iterator();
			//	while(ico.hasNext()){
			//		Renderable r = ico.next();
			//		if(layer.getShapeToObjectMap().get(r) == this){
						
			//		}
			//	}
				if(icon==null){
					icon = new UserFacingIcon("../Cities/icons/houses_icon.gif",
			              new Position(Angle.fromDegreesLatitude(this.getX()),
			            		  Angle.fromDegrees(this.getY()), this.getZ()));
				}
				IconLayer layer2 = new IconLayer();
	           
	            if(this.size<=10)
	            	icon.setSize(new Dimension(15, 15));
	            else if(this.size<=20)
	            	icon.setSize(new Dimension(20, 20));
	            else if(this.size<=40)
	            	icon.setSize(new Dimension(25, 25));
	            else if(this.size<=70)
	            	icon.setSize(new Dimension(30, 30));
	            else if(this.size<=1000)
	            	icon.setSize(new Dimension(35, 35));
	            else
	            	icon.setSize(new Dimension(50, 50));
	            
	            layer2.addIcon(icon);
	            if(iconLayer!=null)
	            	gis.getWwglCanvas().getModel().getLayers().remove(iconLayer);
				gis.getWwglCanvas().getModel().getLayers().add(layer2);
				this.iconLayer=layer2;
//				 canvas.redraw();		 
			 
		 }
	 }
   }
	
	/**
	 * Method to update flow between cities color.
	 */
	@ScheduledMethod(start = 1, interval=10,pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
	public void updateV(){
		 Network network=(Network) RunState.getInstance().getMasterContext().getProjection("Flow");
		 Iterator<RepastEdge> ci = (Iterator<RepastEdge>) network.getEdges(this).iterator();
		 
		 while(ci.hasNext()){
			 RepastEdge e = ci.next();
			 double flow=e.getWeight();
			 Color c=Color.BLUE;
			 if(flow < 10) 
				 c=Color.CYAN;
			 if(flow < 20)
				 c=Color.GREEN;
			 if(flow < 40)
				 c=Color.MAGENTA;
			 if(flow < 60)
				 c=Color.ORANGE;
			 if(flow < 80)
				 c = Color.RED;
			 else
				 c= Color.BLACK;  
		     
			 Renderable re=null;
			 
			 
			 City s = (City)e.getSource();
			 City t = (City)e.getTarget();
		     List<LatLon> latLonList = new ArrayList<LatLon>();
			 latLonList.add(LatLon.fromDegrees(s.getX(),s.getY()));
		     latLonList.add(LatLon.fromDegrees(t.getX(),t.getY()));
			 re = new Polyline(latLonList,100+Math.max(s.getZ(), t.getZ()));
	//		 this.getShapes().put(e, re);
			 
			 ((Polyline)re).setColor(c);
			 List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
			 Iterator<IDisplay> ic = displays.iterator();
			 while(ic.hasNext()){
				 IDisplay display = ic.next();
				 if(display instanceof NewDisplayGIS){
					 if(gis==null)
						 gis = (NewDisplayGIS)display;
					 WorldWindowGLCanvas canvas=gis.getWwglCanvas();
					 
					RenderableLayer rl = new RenderableLayer();			 
					if(oldRend.containsKey(e)){
						  RenderableLayer rll = oldRend.get(e);
						  canvas.getModel().getLayers().remove(rll);  
					}
						
					rl.addRenderable(re);
					canvas.getModel().getLayers().add(rl);
					oldRend.put(e, rl);
				 }
			 }
		}
	}
	
	@ScheduledMethod(start = 5, interval=10,pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
	public void removeV(){
		Network network=(Network) RunState.getInstance().getMasterContext().getProjection("Flow");
		 Iterator<RepastEdge> ci = (Iterator<RepastEdge>) network.getEdges(this).iterator();
		 
		 while(ci.hasNext()){
			 RepastEdge e = ci.next();
			 List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
			 Iterator<IDisplay> ic = displays.iterator();
			 while(ic.hasNext()){
				 IDisplay display = ic.next();
				 if(display instanceof NewDisplayGIS){
					 if(gis==null)
					 gis = (NewDisplayGIS)display;
					 WorldWindowGLCanvas canvas=gis.getWwglCanvas();
				
					  RenderableLayer rll = oldRend.get(e);
					  canvas.getModel().getLayers().remove(rll);  
				 }
			 }
		 }
	}
	
	public void setGeometry(Point point){
		this.point=point;
	}
	
	public Geometry getGeometry(){
		return point;
	}


	public Map<String, Double> getCosts() {
		return costs;
	}


	public void setCosts(Map<String, Double> costs) {
		this.costs = costs;
	}


	public double getVolumeFromI() {
		return volumeFromI;
	}


	public void setVolumeFromI(double volumeFromI) {
		this.volumeFromI = volumeFromI;
	}
	
	
	
}
