package ucl.cities.entities;

import gov.nasa.worldwind.awt.WorldWindowGLCanvas;
import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.layers.RenderableLayer;
import gov.nasa.worldwind.render.Polyline;
import gov.nasa.worldwind.render.Renderable;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.gisnew.AbstractSurfaceLayer;
import repast.simphony.visualization.gisnew.NewDisplayGIS;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;


/**
 * A general location object.
 * 
 * @author maltaweel
 *
 */
public class Location {

	public double x;
	public double y;
	public double z;
	public double origX;
	public double origY;
	public Geometry geo;
	public double gn = Double.POSITIVE_INFINITY;;
	public double fn=Double.POSITIVE_INFINITY;
	public double hn=Double.POSITIVE_INFINITY;
	public Location parent;
	public String name;
	public Geometry geometry;
	public Map<RepastEdge,Renderable>shapes 
 		= new HashMap<RepastEdge,Renderable>();
	public NewDisplayGIS gis;  //gis
	public Point point;
	
	public Location(){
		
	}
	
	public Location(double x, double y, double z){
		this.x=x;
		this.y=y;
		this.z=z;
	}
	
	
	
	public Geometry getGeo() {
		return geo;
	}

	public void setGeo(Geometry geo) {
		this.geo = geo;
	}

	public Geometry getGeometry() {
		return geometry;
	}

	public void setGeometry(Geometry geometry) {
		this.geometry = geometry;
	}

	public double getFn() {
		return fn;
	}

	public void setFn(double fn) {
		this.fn = fn;
	}

	public double getHn() {
		return hn;
	}

	public void setHn(double hn) {
		this.hn = hn;
	}

	public Location getParent() {
		return parent;
	}

	public void setParent(Location parent) {
		this.parent = parent;
	}

	public double getGn() {
		return gn;
	}

	public void setGn(double gn) {
		this.gn = gn;
	}

	public double getX() {
		return x;
	}
	public void setX(double x) {
		this.x = x;
	}
	public double getY() {
		return y;
	}
	public void setY(double y) {
		this.y = y;
	}
	public double getZ() {
		return z;
	}
	public void setZ(double z) {
		this.z = z;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	
	
	
	public double getOrigX() {
		return origX;
	}

	public void setOrigX(double origX) {
		this.origX = origX;
	}

	public double getOrigY() {
		return origY;
	}

	public void setOrigY(double origY) {
		this.origY = origY;
	}

	@ScheduledMethod(start = 0, pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
	public void initV(){
		 Network network=(Network) RunState.getInstance().getMasterContext().getProjection("Canals");
		 Iterator<RepastEdge> ci = (Iterator<RepastEdge>) network.getEdges(this).iterator();
		 
		 while(ci.hasNext()){
			 RepastEdge e = ci.next();
			 Color c=Color.BLUE;
			 
			 Renderable re=null;
			 if(this.getShapes().containsKey(e))
			   re=this.getShapes().get(e);
			 else{
				 Location s = (Location)e.getSource();
				 Location t = (Location)e.getTarget();
				 List<LatLon> latLonList = new ArrayList<LatLon>();
				 latLonList.add(LatLon.fromDegrees(s.getX(),s.getY()));
				 latLonList.add(LatLon.fromDegrees(t.getX(),t.getY()));
				 re = new Polyline(latLonList,100+Math.max(s.getZ(), t.getZ()));
				 this.getShapes().put(e, re);
			 }
			
			 ((Polyline)re).setColor(c);
			 ((Polyline)re).setLineWidth(2.0);
			 List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
			 Iterator<IDisplay> ic = displays.iterator();
			 while(ic.hasNext()){
				 IDisplay display = ic.next();
				 if(display instanceof NewDisplayGIS){
					 if(gis==null)
						 gis = (NewDisplayGIS)display;
					 WorldWindowGLCanvas canvas=gis.getWwglCanvas();
					 	
				     RenderableLayer rl = new RenderableLayer();
				     rl.addRenderable(re);
					 canvas.getModel().getLayers().add(rl);
				 }
			 }
		}
	}
	
	public Map<RepastEdge, Renderable> getShapes() {
		return shapes;
	}


	public void setShapes(Map<RepastEdge, Renderable> shapes) {
		this.shapes = shapes;
	}
	
}
