package ucl.cities.model;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.velocity.runtime.RuntimeInstance;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunInfo;
import repast.simphony.engine.environment.RunState;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;
import ucl.cities.context.CitiesContextBuilder;
import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.City;
import ucl.cities.utilities.DistanceMeasure;

/**
 * Modified Wilson model based on UCL Grant work.
 * 
 * @author Mark Altaweel
 *
 */
public class NewWilsonModel extends WilsonAgentModel{
	
//	Map<City,City>mapFlow=new HashMap<City,City>();
	static Map<Double,Double>tickValue=new HashMap<Double,Double>(); //tick value and total size value for sites.
	
	@Override
	public void step(City c) {
		flow(c);
		
	}
	
	@Override
	public void flow(City c) {
		Context context=RunState.getSafeMasterContext();
		int cSize = CitiesContextBuilder.cities.size();
		
		Iterator<City> ci=context.getObjects(City.class).iterator();
		Network flowN = (Network) RunState.getInstance().getMasterContext().getProjection("Flow");
		Network canals = (Network) RunState.getInstance().getMasterContext().getProjection("Canals");
//		Network ndE = (Network) RunState.getInstance().getMasterContext().getProjection("nd");
		
		while(ci.hasNext()){
			City c2 = ci.next();
			if(c==c2){
	//			distance=0.001;
				continue;
			}
			double distance=0;
			distance=Math.log(canals.getEdge(c,c2).getWeight());
	//		double ccc=((CanalEdge)canals.getEdge(c,c2)).getCrossingWeight();
			
	//		double betaCal=Math.exp(-c.getBeta()*distance);
	//		double cL=Math.exp(-c.getBeta()*distance);
	//		double dl=Math.pow(c2.getSize(),c2.getAlpha());
			double alphaCal=Math.pow(c2.getSize(),c2.getAlpha()*c2.externalFlowPercentage)*Math.exp(-c.getBeta()*distance);
			double betaCal=totalSize(c);
			double numDe=alphaCal/betaCal;
			
		//	double impedence=1/(alphaCal*betaCal);
			double flow=c.getPopulation()*numDe;
			
//		if(c2.getName().equals("Bakr 'Awa")){
//				System.out.println("stop");
//			}
//			if(flow<0)
//				flow=0.0d;
	//		double flow=impedence*c.getPopulation()*alphaCal*betaCal;

			String x=c.getName();
			String y=c2.getName();
			
			String name=x+"-"+y;
			
			RepastEdge e=null;
			
			if(flowEdges.containsKey(name)){
	//			e=flowEdges.get(name);
				e=flowN.getEdge(c, c2);
				if(e==null){
					e = new RepastEdge(c,c2,true,flow);
					flowN.addEdge(e);
					flowEdges.put(name,e);
				}
				else
					e.setWeight(flow);
			}
			else{
				e = new RepastEdge(c,c2,true,flow);
		//		context.add(e);
				flowN.addEdge(e);
				flowEdges.put(name,e);
			}
			((CanalEdge)canals.getEdge(c,c2)).setUsed(flow);
	//		if(c2.getName().contains("Bakr 'Awa") ||c2.getName().contains("Yasin Tape (SSP-002)"))
	//			flow+=100.0;
			
			c2.setFlowValue(flow+(c2.getFlowValue()));
	//		if(Double.toString(c2.getFlowValue()).equals("NaN")){
	//			System.out.println("stop");
	//		}
		}	
//		if(c.getName().equals("Bakr 'Awa")){
//			System.out.println("stop");
//		}
	}
	
	/**
	 * Method to calculate city size based on Harris and Wilson (1978).
	 * @param c a city object.
	 */
	public void calculateSize(City c){
//		if(c.getName().equals("Bakr 'Awa")){
//			System.out.println("stop");
//		}
		double revenue=c.getFlowValue();
		double size=c.getSize()+Epsilon1*(revenue-Constant*c.getSize());
		
		c.setSize(size);
	}
	
	/**
	 * This is done as needed for total size of cities
	 * @return
	 */
	public double totalSize(City c){
		double totalSize=0;
		Network canals = (Network) RunState.getInstance().getMasterContext().getProjection("Canals");
		CitiesContextBuilder.cities.remove(c);
		Iterator<City> ic = CitiesContextBuilder.cities.iterator();
		while(ic.hasNext()){
			City c2 = ic.next();
			double distance=Math.log(canals.getEdge(c,c2).getWeight());
			
			totalSize+=Math.pow(c2.getSize(),c2.alpha*c2.externalFlowPercentage)*Math.exp(-c2.beta*distance);
		}
		CitiesContextBuilder.cities.add(c);
		return totalSize;
	}
	
	/**
	 * Determines flow coming out of c(i).
	 * @param c
	 */
	public void findVolumeofFlow(City c){
		double volumeOfFlow=0;
		
		double totalSize=0;
		if(tickValue.containsKey(RunEnvironment.getInstance().getCurrentSchedule().getTickCount())){
			totalSize=tickValue.get(RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
		}
		else{
			totalSize=totalSize(c);
			tickValue.put(RunEnvironment.getInstance().getCurrentSchedule().getTickCount(),totalSize);
		}
		volumeOfFlow=(CitiesContextBuilder.cities.size()*c.getSize())/totalSize;
		c.setVolumeFromI(volumeOfFlow);
		
	}
}
