package ucl.cities.output;

	import java.io.File;
	import java.util.ArrayList;
	import java.util.Collection;
	import java.util.Iterator;
	import java.util.List;

	import org.geotools.data.DataUtilities;
	import org.geotools.feature.Feature;
	import org.geotools.feature.FeatureCollection;
	import org.geotools.feature.FeatureCollections;
	import org.geotools.feature.FeatureType;
	import org.geotools.feature.IllegalAttributeException;
	import org.geotools.feature.SchemaException;

	import com.vividsolutions.jts.geom.Coordinate;
	import com.vividsolutions.jts.geom.CoordinateSequence;
	import com.vividsolutions.jts.geom.Geometry;
	import com.vividsolutions.jts.geom.GeometryFactory;
	import com.vividsolutions.jts.geom.LineString;
	import com.vividsolutions.jts.geom.MultiLineString;
	import com.vividsolutions.jts.geom.Point;
	import com.vividsolutions.jts.geom.PrecisionModel;

	import repast.simphony.engine.environment.RunEnvironment;
	import repast.simphony.engine.environment.RunState;
	import repast.simphony.engine.schedule.ScheduleParameters;
	import repast.simphony.engine.schedule.ScheduledMethod;
	import repast.simphony.space.graph.Network;
	import repast.simphony.space.graph.RepastEdge;
	import ucl.cities.entities.CanalEdge;
	import ucl.cities.entities.City;
	import ucl.cities.entities.Location;
	import ucl.cities.model.Model;
	import ucl.cities.model.WilsonAgentModel;

	public class CityOutputter<T> {
		
		@ScheduledMethod(start = 10, interval=10,pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
		public void doOutput(){
			Iterator<City> ci=
					RunState.getInstance().getMasterContext().getAgentLayer(City.class).iterator();
			
			Iterator<CanalEdge> cii =  
					((Network) RunState.getInstance().getMasterContext().getProjection("Canals")).getEdges().iterator();
			
			List<City> cities = new ArrayList<City>();
			List<CanalEdge> canals = new ArrayList<CanalEdge>();
			while(ci.hasNext()){
				cities.add((City) ci.next());
			}
			
			while(cii.hasNext()){
				canals.add(cii.next());
			}
			writeShapefileRepeat(cities);
			writeShapefileRepeatCanals(canals);
		}
		
		@ScheduledMethod(start = 1, interval=1,pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
		public void recalculatePopulation(){
			Iterator<City> ci=
					RunState.getInstance().getMasterContext().getAgentLayer(City.class).iterator();
			List<City> cities = new ArrayList<City>();
			double totalFlow=0;
			int totalPopulation=0;
			while(ci.hasNext()){
				City c = ci.next();
				totalFlow+=c.getFlowValue();
				totalPopulation+=c.getPopulation();
			}
			ci=RunState.getInstance().getMasterContext().getAgentLayer(City.class).iterator();
			while(ci.hasNext()){
				City c = ci.next();
				c.setPopulation((int)Math.rint(totalPopulation*c.getFlowValue()/totalFlow));
			}
		}
		
		@ScheduledMethod(start = 10, interval=10, pick=ScheduledMethod.ALL, priority = ScheduleParameters.LAST_PRIORITY)
		public void writeLines(){
			writeNetwork();
		}

		public void writeShapefileRepeat(List<City> cities){
			double tick=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			try {
				writeShapefile(cities, "cities"+Math.rint(tick));
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		public void writeShapefileRepeatCanals(List<CanalEdge> canals){
			double tick=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			try {
				writeShapefileCanals(canals, "canals"+Math.rint(tick));
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		/**
		 * This is now Nystuen-Dacey graph
		 */
		public void writeNetwork(){
			double tick=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			Network flowN = (Network) RunState.getInstance().getMasterContext().getProjection("Flow");
	//		Iterator<City> ic = RunState.getInstance().getMasterContext().getObjects(City.class).iterator();
			Network canals = (Network) RunState.getInstance().getMasterContext().getProjection("Canals");
			FeatureType featureType=null;
            try {
	            featureType = DataUtilities.createType("ConnectionOutput", 
	            		"geom:MultiLineString,Name1:java.lang.String,Name2:java.lang.String,Flow:java.lang.Double,Alpha:java.lang.Double,Beta:java.lang.Double");
            } catch (SchemaException e2) {
	            // TODO Auto-generated catch block
	            e2.printStackTrace();
            }
			FeatureCollection collection = FeatureCollections.newCollection();
			Iterator<RepastEdge> ic = flowN.getEdges().iterator();
			double aveFlowCost=0;
			double graphLineLength=0;
			double cc=0;
			double ll=0;
			while(ic.hasNext()){
				RepastEdge e = ic.next();
				City c =(City) e.getTarget();
				
				City c2 = (City)e.getSource();
				if(c2==c)
					continue;
				Model m = c.getModel();
				RepastEdge e2 = ((WilsonAgentModel)m).getFlowEdges().get(c.getName()+"-"+c2.getName());
				aveFlowCost=+e.getWeight();
				cc++;
				
				if(e.getWeight()>e2.getWeight() && Math.rint(e.getWeight())>0){
					try {
						graphLineLength+=lineCreation(((City)e.getSource()).getPoint(),((City)e.getTarget()).getPoint(),e.getWeight()
								,collection,featureType,c2.getName(),c.getName(),c.alpha,c.beta);
						ll++;
					} catch (SchemaException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					} catch (IllegalAttributeException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
				}
			}
			aveFlowCost=aveFlowCost/cc;  //average flow and distance ratio
			graphLineLength=graphLineLength/ll;  //average Nystuen-Dacey line length
			System.out.println("Average Line Length: "+graphLineLength);
			ShapefileWriter writer = new ShapefileWriter();
			int n=RunState.getInstance().getRunInfo().getRunNumber();
			File shpFile = new File("../Cities/output/","ConnectionOutput"+tick+"_"+n+".shp");
			writer.write(shpFile, collection, featureType);
		}
		
		  public static void writeShapefile(List<City> poly,String category) throws Exception {
			  int n=RunState.getInstance().getRunInfo().getRunNumber();
			    File shpFile = new File("../Cities/output/","CityOutput"+category+"_"+n+".shp");	
			    
				ShapefileWriter writer = new ShapefileWriter();
				FeatureType featureType=DataUtilities.createType("CityOutput", 
						"geom:Point,Tick:java.lang.Double,Name:java.lang.String,Flow:java.lang.Double,Size:java.lang.Double,Population:java.lang.Integer," +
						"Alpha:java.lang.Double,Beta:java.lang.Double");
				FeatureCollection collection = createCollection(poly,featureType);
				writer.write(shpFile,collection,featureType);
			   
			 }
		  
		  public static void writeShapefileCanals(List<CanalEdge> poly,String category) throws Exception {
			   
			    File shpFile = new File("../Cities/output/","CanalsOutput"+category+".shp");	
			    
				ShapefileWriter writer = new ShapefileWriter();
				FeatureType featureType=DataUtilities.createType("CanalsOutput", 
						"geom:MultiLineString,Tick:java.lang.Double,Weight:java.lang.Double,rWeight:java.lang.Double,Count:java.lang.Double");
			//	FeatureCollection collection = createCollectionCanals(poly,featureType);
			//	writer.write(shpFile,collection,featureType);
			   
			 }

		  /**
		   * Method to create a feature collection of structure objects in shulgi
		   * @param Citys the structures
		   * @param featureType the feature type containing the attributes for a shapefile
		   * @return a feature collection
		   * @throws Exception an exception with the data
		   */
		public static FeatureCollection createCollection(List<City> Citys, FeatureType
				featureType) throws Exception{
			 FeatureCollection collection = FeatureCollections.newCollection();
			 double tick = RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			Iterator<City> ic = Citys.iterator();
			while(ic.hasNext()){
				  City f = ic.next();
				  GeometryFactory ff = new GeometryFactory();
				  Point p=ff.createPoint(new Coordinate(f.getOrigX(),f.getOrigY(),f.getZ()));
				  String name=f.getName();
				  double flow=f.getFlowValue();
				  double size=f.getSize();
				  int population=f.getPopulation();
				  Feature feature = featureType.create(new Object[] {p,tick,name,flow,size,population,f.alpha,f.beta});
				  collection.add( feature );
			  }
			 return collection;
		}
		
		public static FeatureCollection createCollectionCanals(List<CanalEdge> canals, FeatureType
				featureType) throws Exception{
			 FeatureCollection collection = FeatureCollections.newCollection();
			 double tick = RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			Iterator<CanalEdge> ic = canals.iterator();
			while(ic.hasNext()){
				  CanalEdge e= ic.next();
				 
				  Location l1=(Location) e.getSource();
				  Location l2=(Location) e.getTarget();
				  
				  Coordinate[] cs = {l1.point.getCoordinate(),l2.point.getCoordinate()};
					
				  LineString l = new LineString(cs,new PrecisionModel(),1);
				  MultiLineString ls = new MultiLineString(new LineString[]{l},new PrecisionModel(),1);
					
				  double weight=e.getWeight();
				  double rWeight=e.getrWeight();
				  double count=e.getUsed();
				  
				  
				  Feature feature = featureType.create(new Object[] {ls,tick,weight,rWeight,count});
				  collection.add( feature );
			  }
			 return collection;
		}
		
		public static double lineCreation(Point p1,Point p2,double weight,FeatureCollection collection,FeatureType featureType,String name1,String name2, double alpha,double beta) 
				throws SchemaException, IllegalAttributeException{
			ShapefileWriter writer = new ShapefileWriter();
			Coordinate[] cs = {p1.getCoordinate(),p2.getCoordinate()};
			
			LineString l = new LineString(cs,new PrecisionModel(),1);
			MultiLineString ls = new MultiLineString(new LineString[]{l},new PrecisionModel(),1);
			
			Feature feature = featureType.create(new Object[] {ls,name1,name2,weight,alpha,beta});
			collection.add( feature );
			return ls.getLength();
		}
	}

