package ucl.cities.output;

import org.geotools.data.DataStoreFactorySpi;
import org.geotools.data.DefaultTransaction;
import org.geotools.data.FeatureStore;
import org.geotools.data.Transaction;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.data.shapefile.ShapefileDataStoreFactory;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureType;
import org.geotools.referencing.crs.DefaultGeographicCRS;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

/**
 * Class to write out data into a shapefile
 * @author 
 */

public class ShapefileWriter {

  /**
  * General constructor
  */
  public ShapefileWriter() {

  }

  /**
   * Method for writing output of a feature collection
   * @param file the file to write to
   * @param collection the feature collection to write out
   * @param featureType the feature type containing attributes
   */
  public void write(File file, FeatureCollection collection, FeatureType featureType) {

	  try {
		    DataStoreFactorySpi factory = new ShapefileDataStoreFactory();
		    Map<String,Object> create = new HashMap<String,Object>();
			
			create.put("url", file.toURI().toURL() );
			create.put("create spatial index",Boolean.TRUE);
			
			ShapefileDataStore newDataStore = (ShapefileDataStore) factory.createNewDataStore( create );
			newDataStore.createSchema( featureType );
			newDataStore.forceSchemaCRS( DefaultGeographicCRS.WGS84 );

		    writeToShapefile(newDataStore,collection);
		  } catch (IOException e) {
			  e.printStackTrace();
		  } 
  }
  
  /**
   * Method called within this class to write out the output
   * @param data the shapefile data stored
   * @param collection the collection of features to write out
   * @throws IOException the file exception if problems occur
   */
  private void writeToShapefile(ShapefileDataStore data, FeatureCollection collection) throws IOException{
	  	Transaction transaction = new DefaultTransaction("create");
		String typeName = data.getTypeNames()[0];
		FeatureStore featureStore = (FeatureStore) data.getFeatureSource( typeName );
		featureStore.setTransaction(transaction);
		
		try {			
			featureStore.addFeatures(collection);
			transaction.commit();
		}
		catch (Exception problem){
			problem.printStackTrace();
			transaction.rollback();
		}
		finally {
			transaction.close();
		}
	}
}


