package ucl.cities.utilities;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Comparator;
import java.util.PriorityQueue;

import repast.simphony.space.graph.Network;
import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.Location;

import edu.uci.ics.jung.graph.Graph;


/**
 * Returns the shortest path as a list of edges
 * See Astar.java(code Mark sent to me) and Wikipedia for further details
 * @author Yanwei Wu and Mark Altaweel
 *
 */

public class AstarShortestPath {

	Network network;
//	ArrayList<CanalEdge> pathList;
	double leastCost;

	public AstarShortestPath (Network network){
		this.network = network;
		this.leastCost = 0;
	}	

	
	/**
	 * @param o1 node1
	 * @param o2 node2
	 * @return the direct distance
	 */
	public double Heuristic(Location o1, Location o2){
		return Math.sqrt(Math.pow(o1.getX()-o2.getX(), 2) + 
				Math.pow(o1.getY()-o2.getY(), 2) + 
				Math.pow(o1.getZ()-o2.getZ(), 2));
	}
		
	/**
	 * @param source source node
	 * @param target target node
	 * @return the path list
	 */
	public ArrayList<CanalEdge> getPath(Object source, Object target){
		ArrayList<CanalEdge> pathList = new ArrayList<CanalEdge>();
		
		Location sourceVertex = (Location)(source);
		Location targetVertex = (Location)(target);

		List settled = new ArrayList();
		
		PriorityQueue<Location> unsettled = new PriorityQueue<Location>( 11,
				new Comparator<Location>() {
					public int compare(Location o1, Location o2) {
						return o1.getFn() < o2.getFn() ? -1 : o1.getFn() == o2.getFn() ? 0 : 1;
					}
				});
		
		sourceVertex.setGn(0);// Gn: shortest path from source to this node
		sourceVertex.setHn(Heuristic(sourceVertex, targetVertex));// may need not to set it
		sourceVertex.setFn(sourceVertex.getGn()+sourceVertex.getHn());// never choose this source to the path again
		sourceVertex.setParent(null);
		
			unsettled.add(sourceVertex);
			while (!unsettled.isEmpty()) {
				Location u = unsettled.poll();
				if (settled.contains(u)) continue; 
				if (u == targetVertex) break;
				for (Object edge0 : network.getEdges(u)){
					CanalEdge<Location> edge = (CanalEdge<Location>)(edge0);
					Location v = (Location) edge.getTarget();
					double weight = edge.getWeight();
					if (v.equals(u))
					{
						if (!network.isDirected())
						{
							v = (Location) edge.getSource();
							weight = edge.getrWeight();
						}
						else
							continue;
					}

					if (v.getHn() == Double.POSITIVE_INFINITY)
						v.setHn(Heuristic(v, targetVertex));
				
			/* if v.getFn() is less than its currrent value, change it. need to decide if v has been in unsettled alreday.
			 * else v must have been in the unsettled already. 
			 */
					if ((u.getGn() + weight + v.getHn()) < v.getFn())
						//	if ((u.getGn() + edge.getWeight()) < v.getGn())
					{
						if (unsettled.contains(v))
						{
							unsettled.remove(v);
						}
						v.setGn(u.getGn() + weight);
						v.setFn(v.getGn() + v.getHn());
						v.setParent(u);
						unsettled.add(v);
					}
				}
				settled.add(u); 
			}
					
			
			Location tmp = targetVertex;	
			if (tmp.getParent() != null) 
			{
				while(tmp != sourceVertex)
				{
					CanalEdge edge = (CanalEdge) network.getEdge(tmp.getParent(), tmp);
					double weight = edge.getWeight();
					if (edge.getSource().equals(tmp)&& !network.isDirected())
					{
					//	edge = (CanalEdge) network.getEdge(tmp, tmp.getParent());
						weight = edge.getrWeight();
					}
					pathList.add(0, edge);
					edge.setUsedWeight(weight);
					leastCost += weight;
					tmp = tmp.getParent();					
				}				
			}
			
			//set the value of leastCost in case the path size is zero
			if (pathList.size() == 0)
				leastCost = Double.POSITIVE_INFINITY;
			resetNode(targetVertex);
			//resetQueue(unsettled);
			resetSet(unsettled, settled);
			return pathList;
	}
	
	public void resetNode(Location n)
	{
		n.setFn(Double.POSITIVE_INFINITY);
		n.setGn(Double.POSITIVE_INFINITY);
		n.setHn(Double.POSITIVE_INFINITY);
		n.setParent(null);
	}
	public void resetSet(PriorityQueue<Location> unsettled, List settled)
	{
		//resetQueue(unsettled);
		while (!unsettled.isEmpty())
		{
			Location v = unsettled.remove();
			resetNode(v);
		}
		while(!settled.isEmpty())
		{
			Location v = (Location) settled.remove(0);
			resetNode(v);				
		}
	}


	public double getLeastCost() {
		return leastCost;
	}


	public void setLeastCost(double leastCost) {
		this.leastCost = leastCost;
	}

}