package ucl.cities.utilities;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.measure.unit.SI;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;

import org.jscience.geography.coordinates.UTM;

import com.vividsolutions.jts.geom.Point;



import repast.simphony.engine.environment.RunState;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;

import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.City;
import ucl.cities.entities.Location;

/**
 * Class for calculating distance.
 * 
 * @author maltaweel
 *
 */
public class DistanceMeasure {
	
	public static Map<String,Double>DistanceMap=new HashMap<String,Double>();
	public static AstarShortestPath shortestPath;
	public static List<CanalEdge> path;
	
	public DistanceMeasure(){
		
	}
	
	public static void notDonkeyEnergy(Point p1, Point p2, Location el1, Location el2, CanalEdge e){
		 double distance=calculateDistance(p1,p2,el1,el2);
		 double deg=percentGrade(p1,p2,el1, el2);
		 double deg2=percentGrade(p1,p2,el2, el1);
		 double weight=1.0;
		 if(deg>0)
			 weight=1.5;
		 e.setWeight(weight*distance);
		 
		 weight=1.0;
		 if(deg2>0)
			 weight=1.5;
		e.setrWeight(weight*distance);
	}
	
	public static double calculateDonkeyEnergy(Point p1, Point p2, Location el1, Location el2){
		double energy=0;
		
		double defaultM=0.97;
		double energyMove=0.97;
		
		double gradient=percentGrade(p1,p2,el1,el2);
		double distance3D=calculateDistance(p1,p2,el1,el2);
		
		
		double travelM=Math.abs((el1.z-el2.z)/distance3D);
		
		if(gradient>0){
			if(gradient>1.0)
				energyMove=RandomHelper.nextDoubleFromTo(5.0, 10.0);
			if(gradient>5.0)
				energyMove=RandomHelper.nextDoubleFromTo(10.0, 20.0);
			if(gradient>9.0)
				energyMove=RandomHelper.nextDoubleFromTo(30.0, 40.0);
		}
		else if(gradient<0){
			if(gradient<-1.0)
				energyMove=RandomHelper.nextDoubleFromTo(0.0,1.0);
			if(gradient<-5.0)
				energyMove=RandomHelper.nextDoubleFromTo(4.0, 5.0);
			if(gradient<-8.0)
				energyMove=RandomHelper.nextDoubleFromTo(1.0,2.0);
		}
		
		energy=1+(distance3D*(Math.abs(energyMove-defaultM)/Math.sin(travelM)));
		
		return energy;
	}
	
	/**
	 * Calculate the percent slope or gradient
	 * @param beginning the beginning location.
	 * @param end the end location
	 * @return the result.
	 */
	public static double percentGrade (Point p1, Point p2, Location beginning, Location end) {
		double results=0f;
		
		double xBeg=p1.getX();
		double yBeg=p1.getY();
		double zBeg=beginning.getZ();
	
		double xEnd=p2.getX();
		double yEnd=p2.getY();
		double zEnd=end.getZ();
	
		double xDifference=xEnd-xBeg;
		double yDifference=yEnd-yBeg;
		double zDifference=zEnd-zBeg;
	
		double distance=Math.sqrt(Math.pow(xDifference, 2) 
			+ Math.pow(yDifference, 2)+Math.pow(zDifference, 2));

		results=(zDifference/distance)*100;
	
		return results;
	}
	
	/**
	 * Method to calculate distance between cities.
	 * @param c1 the first city.
	 * @param c2 the second city.
	 * @return the distance value.
	 */
	public static double measureDistanceBetweenCities(City c1, City c2){
		double distance=0;
		Network<Location> cNetwork=(Network<Location>) RunState.getInstance().getMasterContext().getProjection("Canals");
		
		if(shortestPath==null)
			shortestPath=new AstarShortestPath(cNetwork);
		
		path=shortestPath.getPath(c1.getRoadPointLocation(), c2.getRoadPointLocation());
		
		Iterator<CanalEdge> ic = path.iterator();
		
		while(ic.hasNext()){
			CanalEdge e = ic.next();
			distance+=e.getUsedWeight();
		}

		return distance;
	}
	
	/**
	 * Method to calculate 3d distance.
	 * @param l the first location.
	 * @param l2 the second location.
	 * @return a double distance value.
	 */
	public static double calculateDistance(Point l, Point l2,Location ll, Location ll2){
		Point3d p3d = new Point3d(l.getX(),l.getY(),ll.getZ());
		double distance=p3d.distance(new Point3d(l2.getX(),l2.getY(),ll2.getZ()));
		
		
		return Math.abs(distance);
	}
	
	public static double calculate2DDistance(Point l, Point l2){
		
		Point2d p2d = new Point2d(l.getX(),l.getY());
		double distance=p2d.distance(new Point2d(l2.getX(),l2.getY()));
		
		
		return Math.abs(distance);
	}
	
	public static double degreeSlopeCalculation (Point p1, Point p2, Location beginning, Location end) {
	
		double results=0f;
		
		
		double xBeg=p1.getX();
		double yBeg=p1.getY();
		double zBeg=beginning.getZ();
	
		double xEnd=p2.getX();
		double yEnd=p2.getY();
		double zEnd=end.getZ();
	
		double xDifference=xEnd-xBeg;
		double yDifference=yEnd-yBeg;
		double zDifference=zEnd-zBeg;
	
		double distance=Math.sqrt(Math.pow(xDifference, 2) 
			+ Math.pow(yDifference, 2)+Math.pow(zDifference, 2));

		double ratio=zDifference/distance;
		results=Math.toDegrees(Math.atan(ratio));

		return results;
	}

}
